@@ -2,34 +2,14 @@
 #define __INHERITED_XS_IMPL_H_
 
 /*
-    av_extend() always gives us at least 4 elements, so don't bother with
-    saving memory for need_cb = false version until this struct grows larger
-*/
-
-struct shared_keys {
-    union {
-        SV* hash_key;
-        SV* storage;
-    };
-    SV* pkg_key;
-    SV* read_cb;
-    SV* write_cb;
-};
-
-enum AccessorTypes {
-    Inherited,
-    InheritedCb,
-    PrivateClass
-};
-
-/*
-    These macroses have the following constraints:
+    These macroses impose the following rules:
         - SP is at the start of the args list
-        - afterwards SP may become invalid, so don't touch it
+        - SP may become invalid afterwards, so don't touch it
         - PL_stack_sp is updated when needed
 
-    The latter may be not that obvious, but it's a result of a callback doing stack work for us.
-    But all non-essential updates are not performed after callbacks.
+    The latter may be not that obvious, but it's a result of a callback doing
+    dirty stack work for us. Note that only essential cleanup is done
+    after call_sv().
 */
 
 #define CALL_READ_CB(result, cb)        \
@@ -89,9 +69,9 @@ CAIXS_opmethod_wrapper(pTHX) {
     HV* stash = NULL;
 
     /*
-        this block isn't required for the 'goto gotcv' case, but skipping it
-        (or swapping those blocks) will make unstealing inside 'goto gotcv' block impossible,
-        thus requiring additional check in the fast case, and subref is a failure anyway most of the times
+        This block isn't required for the 'goto gotcv' case, but skipping it
+        (or swapping those blocks) makes unstealing inside 'goto gotcv' block impossible,
+        thus requiring additional check in the fast case, which is to be avoided.
     */
 #ifndef GV_CACHE_ONLY
     if (LIKELY(self != NULL)) {
@@ -131,7 +111,7 @@ CAIXS_opmethod_wrapper(pTHX) {
             SV* const rmeth = SvRV(meth);
             if (SvTYPE(rmeth) == SVt_PVCV) {
                 cv = (CV*)rmeth;
-                goto gotcv; /* we don't care about the 'stash' var here */
+                goto gotcv; /* We don't care about the 'stash' var here */
             }
         }
 
@@ -152,7 +132,7 @@ CAIXS_opmethod_wrapper(pTHX) {
         OP_UNSTEAL(optype);
     }
 
-    HE* he; /* to allow 'goto' to jump over this */
+    HE* he; /* To allow 'goto' to jump over this */
     if ((he = hv_fetch_ent(stash, meth, 0, hash))) {
         GV* gv = (GV*)(HeVAL(he));
         if (isGV(gv) && GvCV(gv) && (!GvCVGEN(gv) || GvCVGEN(gv) == (PL_sub_generation + HvMROMETA(stash)->cache_gen))) {
@@ -180,7 +160,7 @@ gotcv:
 
     } else {
         /*
-            we could also lift off CAIXS_entersub optimization here, but that's a one-time action,
+            We could also lift off CAIXS_entersub optimization here, but that's a one-time action,
             so let it fail on it's own
         */
         OP_UNSTEAL(optype);
@@ -204,7 +184,7 @@ CAIXS_entersub(pTHX) {
             if (UNLIKELY(SvTYPE(sv) != SVt_PVCV)) OP_UNSTEAL(OP_ENTERSUB);
         }
 
-        /* some older gcc's can't deduce correct function - have to add explicit cast  */
+        /* Some older gcc's can't deduce correct function - have to add explicit cast  */
         if (LIKELY(CvXSUB(sv) == (XSUBADDR_t)&CAIXS_entersub_wrapper<type>)) {
             /*
                 Assert against future XPVCV layout change - as for now, xcv_xsub shares space with xcv_root
@@ -240,7 +220,7 @@ CAIXS_install_entersub(pTHX) {
 
 #ifdef OPTIMIZE_OPMETHOD
         OP* methop = cUNOPx(op)->op_first;
-        if (LIKELY(methop != NULL)) {   /* such op can be created by call_sv(G_METHOD_NAMED) */
+        if (LIKELY(methop != NULL)) {   /* Such op can be created by call_sv(G_METHOD_NAMED) */
             while (methop->op_sibling) { methop = methop->op_sibling; }
 
             if (methop->op_next == op) {
@@ -278,6 +258,66 @@ CAIXS_find_keys(CV* cv) {
     return keys;
 }
 
+inline HV*
+CAIXS_find_stash(pTHX_ SV* self, CV* cv) {
+    HV* stash;
+
+    if (SvROK(self)) {
+        stash = SvSTASH(SvRV(self));
+
+    } else {
+        GV* acc_gv = CvGV(cv);
+        if (UNLIKELY(!acc_gv)) croak("Can't have package accessor in anon sub");
+        stash = GvSTASH(acc_gv);
+
+        const char* stash_name = HvNAME(stash);
+        const char* self_name = SvPV_nolen(self);
+        if (strcmp(stash_name, self_name) != 0) {
+            stash = gv_stashsv(self, GV_ADD);
+            if (UNLIKELY(!stash)) croak("Couldn't get required stash");
+        }
+    }
+
+    return stash;
+}
+
+template <> inline
+void
+CAIXS_accessor<Constructor>(pTHX_ SV** SP, CV* cv, HV* stash) {
+    dAXMARK; dITEMS;
+
+    CAIXS_install_entersub<Constructor>(aTHX);
+    if (UNLIKELY(!items)) croak("Usage: $obj->constructor or __PACKAGE__->constructor");
+
+    PL_stack_sp = ++MARK; /* PUTBACK */
+
+    if (!stash) stash = CAIXS_find_stash(aTHX_ *MARK, cv);
+    SV** ret = MARK++;
+
+    SV* self;
+    if (items == 2 && SvROK(*MARK) && SvTYPE(SvRV(*MARK)) == SVt_PVHV) {
+        self = *MARK;
+
+    } else if ((items & 1) == 0) {
+        croak("Odd number of elements in hash constructor");
+
+    } else {
+        HV* hash = newHV();
+
+        while (MARK < SP) {
+            SV* key = *MARK++;
+            /* Don't bother with retval here, as in pp_anonhash */
+            hv_store_ent(hash, key, newSVsv(*MARK++), 0);
+        }
+
+        self = sv_2mortal(newRV_noinc((SV*)hash));
+    }
+
+    sv_bless(self, stash);
+    *ret = self;
+    return;
+}
+
 template <> inline
 void
 CAIXS_accessor<PrivateClass>(pTHX_ SV** SP, CV* cv, HV* stash) {
@@ -289,7 +329,7 @@ CAIXS_accessor<PrivateClass>(pTHX_ SV** SP, CV* cv, HV* stash) {
     shared_keys* keys = (shared_keys*)CAIXS_find_keys(cv);
 
     if (items > 1) {
-        SP -= items; /* no need in the 'items == 1' case */
+        SP -= items; /* No need in the 'items == 1' case */
 
         sv_setsv(keys->storage, *(SP+2));
         PUSHs(keys->storage);
@@ -302,6 +342,7 @@ CAIXS_accessor<PrivateClass>(pTHX_ SV** SP, CV* cv, HV* stash) {
     }
 }
 
+/* covers type = {Inherited, InheritedCb, ObjectOnly} */
 template <AccessorTypes type> inline
 void
 CAIXS_accessor(pTHX_ SV** SP, CV* cv, HV* stash) {
@@ -334,33 +375,22 @@ CAIXS_accessor(pTHX_ SV** SP, CV* cv, HV* stash) {
             if (hent) {
                 CALL_READ_CB(HeVAL(hent), keys->read_cb);
                 return;
+
+            } else if (type == ObjectOnly) {
+                CALL_READ_CB(&PL_sv_undef, keys->read_cb);
+                return;
             }
         }
     }
 
-    /* Couldn't find value in object, so initiate a package lookup. */
-
-#ifdef OPTIMIZE_OPMETHOD
-    if (!stash) {
-#endif
-    if (SvROK(self)) {
-        stash = SvSTASH(SvRV(self));
-
-    } else {
-        GV* acc_gv = CvGV(cv);
-        if (!acc_gv) croak("Can't have package accessor in anon sub");
-        stash = GvSTASH(acc_gv);
-
-        const char* stash_name = HvNAME(stash);
-        const char* self_name = SvPV_nolen(self);
-        if (strcmp(stash_name, self_name) != 0) {
-            stash = gv_stashsv(self, GV_ADD);
-            if (!stash) croak("Couldn't get required stash");
-        }
+    if (type == ObjectOnly) {
+        croak("Can't use object accessor on non-object");
+        return; /* gcc detects unreachability even with bare croak(), but it won't hurt */
     }
-#ifdef OPTIMIZE_OPMETHOD
-    }
-#endif
+
+    /* Couldn't find value in the object, so initiate a package lookup. */
+
+    if (!stash) stash = CAIXS_find_stash(aTHX_ self, cv);
 
     HE* hent;
     if (items > 1) {
@@ -372,7 +402,7 @@ CAIXS_accessor(pTHX_ SV** SP, CV* cv, HV* stash) {
             gv_init_sv(glob, stash, keys->pkg_key, 0);
 
             if (hent) {
-                /* there was just a stub instead of a full glob */
+                /* There was just a stub instead of the full glob */
                 SvREFCNT_inc_simple_void_NN((SV*)glob);
                 SvREFCNT_dec_NN(HeVAL(hent));
                 HeVAL(hent) = (SV*)glob;
